@@ -40,15 +40,7 @@ module ActiveRecord
 
             constraint = build_constraint(klass, table, key, foreign_table, foreign_key)
 
-            predicate_builder = PredicateBuilder.new(TableMetadata.new(klass, table))
-            scope_chain_items = reflection.join_scopes(table, predicate_builder)
-            klass_scope       = reflection.klass_join_scope(table, predicate_builder)
-
-            scope_chain_items.concat [klass_scope].compact
-
-            rel = scope_chain_items.inject(scope_chain_items.shift) do |left, right|
-              left.merge right
-            end
+            rel = reflection.join_scope(table)
 
             if rel && !rel.arel.constraints.empty?
               binds += rel.bound_attributes
