@@ -90,7 +90,11 @@ module ActiveRecord
           end
 
           def key_conversion_required?
-            @key_conversion_required ||= association_key_type != owner_key_type
+            unless defined?(@key_conversion_required)
+              @key_conversion_required = (association_key_type != owner_key_type)
+            end
+
+            @key_conversion_required
           end
 
           def convert_key(key)
@@ -127,42 +131,15 @@ module ActiveRecord
           end
 
           def build_scope
-            scope = klass.unscoped
+            scope = klass.scope_for_association
 
-            values = reflection_scope.values
-            preload_values = preload_scope.values
-
-            scope.where_clause = reflection_scope.where_clause + preload_scope.where_clause
-            scope.references_values = Array(values[:references]) + Array(preload_values[:references])
-
-            if preload_values[:select] || values[:select]
-              scope._select!(preload_values[:select] || values[:select])
-            end
-            scope.includes! preload_values[:includes] || values[:includes]
-            if preload_scope.joins_values.any?
-              scope.joins!(preload_scope.joins_values)
-            else
-              scope.joins!(reflection_scope.joins_values)
+            if reflection.type
+              scope.where!(reflection.type => model.base_class.sti_name)
             end
 
-            if order_values = preload_values[:order] || values[:order]
-              scope.order!(order_values)
-            end
-
-            if preload_values[:reordering] || values[:reordering]
-              scope.reordering_value = true
-            end
-
-            if preload_values[:readonly] || values[:readonly]
-              scope.readonly!
-            end
-
-            if options[:as]
-              scope.where!(klass.table_name => { reflection.type => model.base_class.sti_name })
-            end
-
-            scope.unscope_values = Array(values[:unscope]) + Array(preload_values[:unscope])
-            klass.scope_for_association.merge(scope)
+            scope.merge!(reflection_scope)
+            scope.merge!(preload_scope) if preload_scope != NULL_RELATION
+            scope
           end
       end
     end
