@@ -5,13 +5,13 @@ module ActiveRecord
         class Uuid < Type::Value # :nodoc:
           ACCEPTABLE_UUID = %r{\A\{?([a-fA-F0-9]{4}-?){8}\}?\z}x
 
-          alias_method :type_cast_for_database, :type_cast_from_database
+          alias_method :serialize, :deserialize
 
           def type
             :uuid
           end
 
-          def type_cast(value)
+          def cast(value)
             value.to_s[ACCEPTABLE_UUID, 0]
           end
         end
